# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import unittest

from orquesta import exceptions as exc
from orquesta.specs import types as spec_types
from orquesta.tests.unit.specs import base as test_specs
from orquesta.utils import jsonify as json_util


class SpecTest(unittest.TestCase):
    def setUp(self):
        super(SpecTest, self).setUp()
        self.maxDiff = None

    def test_get_catalog(self):
        self.assertEqual(test_specs.MockSpec.get_catalog(), "test")
        self.assertEqual(test_specs.MockMappingSpec.get_catalog(), "test")
        self.assertEqual(test_specs.MockSequenceSpec.get_catalog(), "test")

    def test_get_version(self):
        self.assertEqual("1.0", test_specs.MockSpec.get_version())

    def test_get_schema(self):
        schema = {
            "type": "object",
            "properties": {
                "name": spec_types.NONEMPTY_STRING,
                "version": spec_types.VERSION,
                "description": spec_types.NONEMPTY_STRING,
                "tags": spec_types.UNIQUE_STRING_LIST,
                "inputs": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
                "vars": spec_types.NONEMPTY_DICT,
                "attr1": spec_types.NONEMPTY_STRING,
                "attr1-1": spec_types.NONEMPTY_STRING,
                "attr1_2": spec_types.NONEMPTY_STRING,
                "attr2": spec_types.NONEMPTY_DICT,
                "attr3": spec_types.UNIQUE_STRING_LIST,
                "attr4": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
                "attr5": test_specs.MockJointSpec.get_schema(includes=None),
                "attr6": test_specs.MockMappingSpec.get_schema(includes=None),
                "attr7": test_specs.MockSequenceSpec.get_schema(includes=None),
            },
            "required": ["attr1"],
            "additionalProperties": False,
        }

        self.assertDictEqual(schema, test_specs.MockSpec.get_schema())

    def test_get_schema_specs_not_resolved(self):
        schema = {
            "type": "object",
            "properties": {
                "name": spec_types.NONEMPTY_STRING,
                "version": spec_types.VERSION,
                "description": spec_types.NONEMPTY_STRING,
                "tags": spec_types.UNIQUE_STRING_LIST,
                "inputs": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
                "vars": spec_types.NONEMPTY_DICT,
                "attr1": spec_types.NONEMPTY_STRING,
                "attr1-1": spec_types.NONEMPTY_STRING,
                "attr1_2": spec_types.NONEMPTY_STRING,
                "attr2": spec_types.NONEMPTY_DICT,
                "attr3": spec_types.UNIQUE_STRING_LIST,
                "attr4": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
                "attr5": test_specs.MockJointSpec,
                "attr6": test_specs.MockMappingSpec,
                "attr7": test_specs.MockSequenceSpec,
            },
            "required": ["attr1"],
            "additionalProperties": False,
        }

        self.assertDictEqual(schema, test_specs.MockSpec.get_schema(resolve_specs=False))

    def test_get_schema_no_meta(self):
        schema = {
            "type": "object",
            "properties": {
                "inputs": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
                "vars": spec_types.NONEMPTY_DICT,
                "attr1": spec_types.NONEMPTY_STRING,
                "attr1-1": spec_types.NONEMPTY_STRING,
                "attr1_2": spec_types.NONEMPTY_STRING,
                "attr2": spec_types.NONEMPTY_DICT,
                "attr3": spec_types.UNIQUE_STRING_LIST,
                "attr4": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
                "attr5": test_specs.MockJointSpec.get_schema(includes=None),
                "attr6": test_specs.MockMappingSpec.get_schema(includes=None),
                "attr7": test_specs.MockSequenceSpec.get_schema(includes=None),
            },
            "required": ["attr1"],
            "additionalProperties": False,
        }

        self.assertDictEqual(schema, test_specs.MockSpec.get_schema(includes=None))

    def test_instance_schema(self):
        schema = {
            "type": "object",
            "properties": {
                "inputs": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
                "vars": spec_types.NONEMPTY_DICT,
                "attr1": spec_types.NONEMPTY_STRING,
                "attr1-1": spec_types.NONEMPTY_STRING,
                "attr1_2": spec_types.NONEMPTY_STRING,
                "attr2": spec_types.NONEMPTY_DICT,
                "attr3": spec_types.UNIQUE_STRING_LIST,
                "attr4": spec_types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
                "attr5": test_specs.MockJointSpec,
                "attr6": test_specs.MockMappingSpec,
                "attr7": test_specs.MockSequenceSpec,
            },
            "required": ["attr1"],
            "additionalProperties": False,
        }

        self.assertDictEqual(schema, test_specs.MockSpec._schema)

    def test_spec_init_arg_none_type(self):
        self.assertRaises(ValueError, test_specs.MockSpec, None)

    def test_spec_init_arg_empty_str(self):
        self.assertRaises(ValueError, test_specs.MockSpec, "")

    def test_spec_init(self):
        spec = {
            "name": "mock",
            "version": "1.0",
            "description": "This is a mock spec.",
            "inputs": ["x", {"y": "polo"}],
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr1": "foobar",
            "attr1-1": "fubar",
            "attr1_2": "foosball",
            "attr2": {"macro": "polo"},
            "attr3": ["<% ctx().var1 %>"],
            "attr4": [{"open": "sesame"}, {"sesame": "open"}],
            "attr5": {"attr1": {"attr1": "<% ctx().var2 %> <% ctx().var3 %>"}},
            "attr6": {"attr1": {"attr1": {"attr1": "wunderbar"}}},
            "attr7": [{"attr1": {"attr1": "wunderbar"}}, {"attr1": {"attr1": "wonderful"}}],
        }

        spec_obj = test_specs.MockSpec(spec)

        self.assertDictEqual(spec_obj.spec, spec)

        # Test properties from the meta schema.
        self.assertEqual(spec_obj.name, spec["name"])
        self.assertEqual(spec_obj.version, spec["version"])
        self.assertEqual(spec_obj.description, spec["description"])

        # Test simple properties from the schema.
        self.assertEqual(spec_obj.attr1, spec["attr1"])
        self.assertDictEqual(spec_obj.attr2, spec["attr2"])
        self.assertListEqual(spec_obj.attr3, spec["attr3"])

        # Test properties with dash or underscore in property name.
        self.assertEqual(spec_obj.attr1_1, spec["attr1-1"])
        self.assertEqual(getattr(spec_obj, "attr1-1"), spec["attr1-1"])
        self.assertEqual(spec_obj.attr1_2, spec["attr1_2"])
        self.assertIsNone(getattr(spec_obj, "attr1-2", None))

        # Test spec nesting.
        self.assertIsInstance(spec_obj.attr5, test_specs.MockJointSpec)
        self.assertIsInstance(spec_obj.attr5.attr1, test_specs.MockLeafSpec)

        self.assertEqual(spec_obj.attr5.attr1.attr1, spec["attr5"]["attr1"]["attr1"])

        # Test dictionary based spec class.
        self.assertIsInstance(spec_obj.attr6, test_specs.MockMappingSpec)
        self.assertIsInstance(spec_obj.attr6.attr1, test_specs.MockJointSpec)
        self.assertIsInstance(spec_obj.attr6.attr1.attr1, test_specs.MockLeafSpec)

        self.assertEqual(spec_obj.attr6.attr1.attr1.attr1, spec["attr6"]["attr1"]["attr1"]["attr1"])

        attr6_keys = list(spec_obj.attr6.keys())
        attr6_values = list(spec_obj.attr6.values())
        attr6_items = list(spec_obj.attr6.items())
        attr6_iter = [item for item in spec_obj.attr6]
        attr6_iteritems = [(k, v) for k, v in six.iteritems(spec_obj.attr6)]

        self.assertListEqual(attr6_keys, ["attr1"])
        self.assertListEqual(attr6_values, [spec_obj.attr6.attr1])
        self.assertListEqual(attr6_items, [("attr1", spec_obj.attr6.attr1)])
        self.assertListEqual(attr6_iter, [("attr1", spec_obj.attr6.attr1)])
        self.assertListEqual(attr6_iteritems, [("attr1", spec_obj.attr6.attr1)])
        self.assertEqual(spec_obj.attr6["attr1"], spec_obj.attr6.attr1)

        # Test list based spec class.
        self.assertIsInstance(spec_obj.attr7, test_specs.MockSequenceSpec)
        self.assertEqual(len(spec_obj.attr7), 2)
        self.assertIsInstance(spec_obj.attr7[0], test_specs.MockJointSpec)
        self.assertIsInstance(spec_obj.attr7[1], test_specs.MockJointSpec)

        # Test non-existent attribute.
        self.assertRaises(AttributeError, getattr, spec_obj, "attr9")

    def test_spec_serialize(self):
        spec = {
            "name": "mock",
            "version": "1.0",
            "description": "This is a mock spec.",
            "inputs": ["x", {"y": "polo"}],
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr1": "foobar",
            "attr1-1": "fubar",
            "attr1_2": "foosball",
            "attr2": {"macro": "polo"},
            "attr3": ["<% ctx().var1 %>"],
            "attr4": [{"open": "sesame"}, {"sesame": "open"}],
            "attr5": {"attr1": {"attr1": "<% ctx().var2 %> <% ctx().var3 %>"}},
            "attr6": {"attr1": {"attr1": {"attr1": "wunderbar"}}},
            "attr7": [{"attr1": {"attr1": "wunderbar"}}, {"attr1": {"attr1": "wonderful"}}],
        }

        spec_obj_1 = test_specs.MockSpec(spec)

        expected_data = {
            "catalog": spec_obj_1.get_catalog(),
            "version": spec_obj_1.get_version(),
            "name": spec_obj_1.name,
            "spec": spec_obj_1.spec,
        }

        spec_obj_1_json = spec_obj_1.serialize()

        self.assertDictEqual(spec_obj_1_json, expected_data)

        spec_obj_2 = test_specs.MockSpec.deserialize(spec_obj_1_json)

        self.assertEqual(spec_obj_2.name, spec_obj_1.name)
        self.assertEqual(spec_obj_2.member, spec_obj_1.member)
        self.assertDictEqual(spec_obj_2.spec, spec_obj_1.spec)

    def test_spec_deserialize_errors(self):
        spec = {
            "name": "mock",
            "version": "1.0",
            "description": "This is a mock spec.",
            "inputs": ["x", {"y": "polo"}],
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr1": "foobar",
            "attr1-1": "fubar",
            "attr1_2": "foosball",
            "attr2": {"macro": "polo"},
            "attr3": ["<% ctx().var1 %>"],
            "attr4": [{"open": "sesame"}, {"sesame": "open"}],
            "attr5": {"attr1": {"attr1": "<% ctx().var2 %> <% ctx().var3 %>"}},
            "attr6": {"attr1": {"attr1": {"attr1": "wunderbar"}}},
            "attr7": [{"attr1": {"attr1": "wunderbar"}}, {"attr1": {"attr1": "wonderful"}}],
        }

        spec_obj = test_specs.MockSpec(spec)

        spec_obj_json = spec_obj.serialize()

        # Test missing catalog information.
        test_json = json_util.deepcopy(spec_obj_json)
        test_json.pop("catalog")

        self.assertRaises(ValueError, test_specs.MockSpec.deserialize, test_json)

        # Test missing version information.
        test_json = json_util.deepcopy(spec_obj_json)
        test_json.pop("version")

        self.assertRaises(ValueError, test_specs.MockSpec.deserialize, test_json)

        # Test mismatch version information.
        test_json = json_util.deepcopy(spec_obj_json)
        test_json["version"] = "99.99"

        self.assertRaises(ValueError, test_specs.MockSpec.deserialize, test_json)

    def test_spec_init_name_not_given(self):
        spec = {
            "version": "1.0",
            "description": "This is a mock spec.",
            "vars": {"var1": "foobar", "var2": "macro", "var3": "polo"},
            "attr1": "foobar",
        }

        spec_obj = test_specs.MockSpec(spec)

        self.assertDictEqual(spec_obj.spec, spec)
        self.assertIsNone(spec_obj.name)

    def test_spec_init_name_arg_given(self):
        spec = {
            "version": "1.0",
            "description": "This is a mock spec.",
            "vars": {"var1": "foobar", "var2": "macro", "var3": "polo"},
            "attr1": "foobar",
        }

        spec_obj = test_specs.MockSpec(spec, name="mock")

        self.assertDictEqual(spec_obj.spec, spec)
        self.assertEqual(spec_obj.name, "mock")

    def test_spec_init_just_required(self):
        spec = {
            "name": "mock",
            "version": "1.0",
            "description": "This is a mock spec.",
            "vars": {"var1": "foobar", "var2": "macro", "var3": "polo"},
            "attr1": "foobar",
        }

        spec_obj = test_specs.MockSpec(spec)

        self.assertDictEqual(spec_obj.spec, spec)
        self.assertEqual(spec_obj.name, spec["name"])
        self.assertEqual(spec_obj.version, spec["version"])
        self.assertEqual(spec_obj.description, spec["description"])
        self.assertEqual(spec_obj.attr1, spec["attr1"])
        self.assertIsNone(spec_obj.attr2)
        self.assertIsNone(spec_obj.attr3)
        self.assertIsNone(spec_obj.attr4)
        self.assertIsNone(spec_obj.attr5)
        self.assertIsNone(spec_obj.attr6)
        self.assertRaises(AttributeError, getattr, spec_obj, "attr9")

    def test_spec_valid(self):
        spec = {
            "name": "mock",
            "version": "1.0",
            "description": "This is a mock spec.",
            "inputs": ["x", {"y": "polo"}],
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr1": "foobar",
            "attr2": {"macro": "polo"},
            "attr3": ["<% ctx().var1 %>"],
            "attr4": [{"open": "sesame"}, {"sesame": "open"}],
            "attr5": {"attr1": {"attr1": "<% ctx().var2 %> <% ctx().var3 %>"}},
        }

        spec_obj = test_specs.MockSpec(spec)

        self.assertDictEqual(spec_obj.spec, spec)
        self.assertDictEqual(spec_obj.inspect(), {})

    def test_spec_invalid(self):
        spec = {
            "name": "mock",
            "version": None,
            "description": "This is a mock spec.",
            "inputs": ["x", {"y": "polo"}],
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr2": {"macro": "polo"},
            "attr3": ['<% 1 +/ 2 %> and <% {"a": 123} %>'],
            "attr4": [{"var1": "<% ctx().fubar %>"}],
            "attr5": {"attr1": {"attr1": "<% <% ctx().var1 %> %>", "attr2": "<% ctx().foobar %>"}},
        }

        errors = {
            "syntax": [
                {
                    "spec_path": "version",
                    "schema_path": "properties.version.anyOf",
                    "message": "None is not valid under any " "of the given schemas",
                },
                {
                    "spec_path": None,
                    "schema_path": "required",
                    "message": "'attr1' is a required property",
                },
            ],
            "expressions": [
                {
                    "type": "yaql",
                    "expression": "<% 1 +/ 2 %>",
                    "spec_path": "attr3",
                    "schema_path": "properties.attr3",
                    "message": "Parse error: unexpected '/' at "
                    "position 3 of expression '1 +/ 2'",
                },
                {
                    "type": "yaql",
                    "expression": '<% {"a": 123} %>',
                    "spec_path": "attr3",
                    "schema_path": "properties.attr3",
                    "message": "Lexical error: illegal character " "':' at position 4",
                },
                {
                    "type": "yaql",
                    "expression": "<% <% ctx().var1 %>",
                    "spec_path": "attr5.attr1.attr1",
                    "schema_path": ("properties.attr5." "properties.attr1." "properties.attr1"),
                    "message": "Parse error: unexpected '<' at position 0 "
                    "of expression '<% ctx().var1'",
                },
            ],
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().fubar %>",
                    "spec_path": "attr4[0]",
                    "schema_path": "properties.attr4",
                    "message": 'Variable "fubar" is referenced ' "before assignment.",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().foobar %>",
                    "spec_path": "attr5.attr1.attr2",
                    "schema_path": ("properties.attr5." "properties.attr1." "properties.attr2"),
                    "message": 'Variable "foobar" is referenced ' "before assignment.",
                },
            ],
        }

        spec_obj = test_specs.MockSpec(spec)

        self.assertDictEqual(spec_obj.inspect(), errors)

    def test_spec_invalid_raise_exception(self):
        spec = {
            "name": "mock",
            "version": None,
            "description": "This is a mock spec.",
            "inputs": ["x", {"y": "polo"}],
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr2": {"macro": "polo"},
            "attr3": ['<% 1 +/ 2 %> and <% {"a": 123} %>'],
            "attr4": [{"var1": "<% ctx().fubar %>"}],
            "attr5": {"attr1": {"attr1": "<% <% ctx().var1 %> %>", "attr2": "<% ctx().foobar %>"}},
        }

        errors = {
            "syntax": [
                {
                    "spec_path": "version",
                    "schema_path": "properties.version.anyOf",
                    "message": "None is not valid under any " "of the given schemas",
                },
                {
                    "spec_path": None,
                    "schema_path": "required",
                    "message": "'attr1' is a required property",
                },
            ],
            "expressions": [
                {
                    "type": "yaql",
                    "expression": "<% 1 +/ 2 %>",
                    "spec_path": "attr3",
                    "schema_path": "properties.attr3",
                    "message": "Parse error: unexpected '/' at "
                    "position 3 of expression '1 +/ 2'",
                },
                {
                    "type": "yaql",
                    "expression": '<% {"a": 123} %>',
                    "spec_path": "attr3",
                    "schema_path": "properties.attr3",
                    "message": "Lexical error: illegal character " "':' at position 4",
                },
                {
                    "type": "yaql",
                    "expression": "<% <% ctx().var1 %>",
                    "spec_path": "attr5.attr1.attr1",
                    "schema_path": ("properties.attr5." "properties.attr1." "properties.attr1"),
                    "message": "Parse error: unexpected '<' at position 0 "
                    "of expression '<% ctx().var1'",
                },
            ],
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().fubar %>",
                    "spec_path": "attr4[0]",
                    "schema_path": "properties.attr4",
                    "message": 'Variable "fubar" is referenced ' "before assignment.",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().foobar %>",
                    "spec_path": "attr5.attr1.attr2",
                    "schema_path": ("properties.attr5." "properties.attr1." "properties.attr2"),
                    "message": 'Variable "foobar" is referenced ' "before assignment.",
                },
            ],
        }

        spec_obj = test_specs.MockSpec(spec)

        with self.assertRaises(exc.WorkflowInspectionError) as ctx:
            spec_obj.inspect(raise_exception=True)

        self.assertEqual(len(ctx.exception.args), 2)
        self.assertEqual(ctx.exception.args[0], "Workflow definition failed inspection.")
        self.assertDictEqual(ctx.exception.args[1], errors)

    def test_spec_valid_with_app_ctx(self):
        spec = {
            "name": "mock",
            "version": "1.0",
            "description": "This is a mock spec.",
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr1": "foobar",
            "attr2": {"macro": "polo"},
            "attr3": ["<% ctx().var1 %>"],
            "attr4": [{"open": "sesame"}, {"sesame": "open"}],
            "attr5": {"attr1": {"attr1": "<% ctx().var2 %> <% ctx().var3 %>"}},
        }

        spec_obj = test_specs.MockSpec(spec)
        app_ctx = {"x": "marco", "y": "polo"}

        self.assertDictEqual(spec_obj.spec, spec)
        self.assertDictEqual(spec_obj.inspect(app_ctx=app_ctx), {})

    def test_spec_invalid_with_bad_app_ctx(self):
        spec = {
            "name": "mock",
            "version": "1.0",
            "description": "This is a mock spec.",
            "vars": {"var1": "foobar", "var2": "<% ctx().x %>", "var3": "<% ctx().y %>"},
            "attr1": "foobar",
            "attr2": {"macro": "polo"},
            "attr3": ["<% ctx().var1 %>"],
            "attr4": [{"open": "sesame"}, {"sesame": "open"}],
            "attr5": {"attr1": {"attr1": "<% ctx().var2 %> <% ctx().var3 %>"}},
        }

        spec_obj = test_specs.MockSpec(spec)
        app_ctx = [{"x": "marco"}, {"y": "polo"}]

        self.assertDictEqual(spec_obj.spec, spec)

        self.assertRaises(TypeError, spec_obj.inspect, app_ctx=app_ctx)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.utils import expression as expr_util


class ExpressionUtilsTest(unittest.TestCase):
    def test_format_error(self):
        expected = {"type": "yaql", "expression": "$.foo in $.bar", "message": "Unknown error."}

        actual = expr_util.format_error("yaql", "$.foo in $.bar", Exception("Unknown error."))

        self.assertEqual(expected, actual)

    def test_format_error_str(self):
        expected = {"type": "yaql", "expression": "$.foo in $.bar", "message": "Unknown error."}

        actual = expr_util.format_error("yaql", "$.foo in $.bar", "Unknown error.")

        self.assertEqual(expected, actual)

    def test_format_error_with_paths(self):
        expected = {
            "type": "yaql",
            "expression": "$.foo in $.bar",
            "spec_path": "path.to.error.in.spec",
            "schema_path": "path.to.reference.in.schema",
            "message": "Unknown error.",
        }

        actual = expr_util.format_error(
            "yaql",
            "$.foo in $.bar",
            Exception("Unknown error."),
            spec_path="path.to.error.in.spec",
            schema_path="path.to.reference.in.schema",
        )

        self.assertEqual(expected, actual)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta import exceptions as exc
from orquesta.utils import schema as schema_util


class SchemaUtilsTest(unittest.TestCase):
    def test_get_schema_type(self):
        self.assertEqual("object", schema_util.get_schema_type({"type": "object"}))
        self.assertIsNone(schema_util.get_schema_type({}))
        self.assertIsNone(schema_util.get_schema_type(None))

    def test_check_schema_mergeable(self):
        schema_util.check_schema_mergeable(None)
        schema_util.check_schema_mergeable({})
        schema_util.check_schema_mergeable({"type": "object"})
        schema_util.check_schema_mergeable({"type": "array"})

    def test_check_schema_unmergeable(self):
        self.assertRaises(
            exc.SchemaIncompatibleError, schema_util.check_schema_mergeable, {"type": None}
        )

        self.assertRaises(
            exc.SchemaIncompatibleError, schema_util.check_schema_mergeable, {"type": "string"}
        )

    def test_check_schemas_compatible(self):
        cases = [{"type": "object"}, {"type": "array"}, {"type": None}, {}, None]

        for s in cases:
            self.assertEqual(
                schema_util.get_schema_type(s), schema_util.check_schemas_compatible(s, s)
            )

    def test_check_schemas_incompatible(self):
        self.assertRaises(
            exc.SchemaIncompatibleError,
            schema_util.check_schemas_compatible,
            {"type": "object"},
            {"type": "array"},
        )

        self.assertRaises(
            exc.SchemaIncompatibleError,
            schema_util.check_schemas_compatible,
            {"type": "object"},
            None,
        )

        self.assertRaises(
            exc.SchemaIncompatibleError,
            schema_util.check_schemas_compatible,
            {"type": "object"},
            {},
        )

    def test_merge_schema_noop(self):
        s = {
            "type": "object",
            "properties": {"name": {"type": "string"}, "description": {"type": "string"}},
        }

        self.assertDictEqual(s, schema_util.merge_schema(s, None))
        self.assertDictEqual(s, schema_util.merge_schema(None, s))
        self.assertDictEqual({}, schema_util.merge_schema(None, None))
        self.assertDictEqual({}, schema_util.merge_schema({}, {}))

    def test_merge_schema_blank(self):
        s = {
            "type": "object",
            "properties": {"name": {"type": "string"}, "description": {"type": "string"}},
        }

        s_blank_obj = {"type": "object"}
        s_blank_arr = {"type": "array"}

        self.assertDictEqual(s, schema_util.merge_schema(s, s_blank_obj))
        self.assertDictEqual(s, schema_util.merge_schema(s_blank_obj, s))
        self.assertDictEqual(s, schema_util.merge_schema(s, s_blank_arr))
        self.assertDictEqual(s, schema_util.merge_schema(s_blank_arr, s))
        self.assertDictEqual({}, schema_util.merge_schema(s_blank_obj, s_blank_obj))
        self.assertDictEqual({}, schema_util.merge_schema(s_blank_arr, s_blank_arr))
        self.assertDictEqual({}, schema_util.merge_schema(s_blank_obj, s_blank_arr))
        self.assertDictEqual({}, schema_util.merge_schema(s_blank_arr, s_blank_obj))
        self.assertDictEqual({}, schema_util.merge_schema(None, s_blank_obj))
        self.assertDictEqual({}, schema_util.merge_schema(s_blank_obj, None))
        self.assertDictEqual({}, schema_util.merge_schema(None, s_blank_arr))
        self.assertDictEqual({}, schema_util.merge_schema(s_blank_arr, None))
        self.assertDictEqual({}, schema_util.merge_schema({}, s_blank_obj))
        self.assertDictEqual({}, schema_util.merge_schema(s_blank_obj, {}))
        self.assertDictEqual({}, schema_util.merge_schema({}, s_blank_arr))
        self.assertDictEqual({}, schema_util.merge_schema(s_blank_arr, {}))

    def test_merge_object_schema(self):
        s1 = {
            "type": "object",
            "properties": {"name": {"type": "string"}, "counter": {"type": "integer"}},
            "required": ["name"],
            "additionalProperties": False,
        }

        s2 = {
            "type": "object",
            "properties": {"description": {"type": "string"}, "counter": {"type": "number"}},
            "required": ["counter"],
            "additionalProperties": True,
        }

        expected = {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "description": {"type": "string"},
                "counter": {"type": "number"},
            },
            "required": ["counter", "name"],
            "additionalProperties": False,
        }

        self.assertDictEqual(expected, schema_util.merge_schema(s1, s2))

    def test_merge_object_schema_overwrite_false(self):
        s1 = {
            "type": "object",
            "properties": {"name": {"type": "string"}, "counter": {"type": "integer"}},
            "required": ["name"],
            "additionalProperties": False,
        }

        s2 = {
            "type": "object",
            "properties": {"description": {"type": "string"}, "counter": {"type": "number"}},
            "required": ["counter"],
            "additionalProperties": True,
        }

        expected = {
            "type": "object",
            "properties": {
                "name": {"type": "string"},
                "description": {"type": "string"},
                "counter": {"type": "integer"},
            },
            "required": ["counter", "name"],
            "additionalProperties": False,
        }

        self.assertDictEqual(expected, schema_util.merge_schema(s1, s2, False))

    def test_merge_object_schema_patterned_properties(self):
        s1 = {
            "type": "object",
            "patternProperties": {"^version$": {"type": "integer"}},
            "minProperties": 0,
            "maxProperties": 10,
        }

        s2 = {
            "type": "object",
            "patternProperties": {
                r"^version$": {"type": "number"},
                r"^(?!version)\w+$": {"type": "string"},
            },
            "minProperties": 1,
            "maxProperties": 100,
        }

        expected = {
            "type": "object",
            "patternProperties": {
                r"^version$": {"type": "number"},
                r"^(?!version)\w+$": {"type": "string"},
            },
            "minProperties": 1,
            "maxProperties": 10,
        }

        self.assertDictEqual(expected, schema_util.merge_schema(s1, s2))

    def test_merge_object_schema_patterned_properties_overwrite_false(self):
        s1 = {
            "type": "object",
            "patternProperties": {"^version$": {"type": "integer"}},
            "minProperties": 1,
            "maxProperties": 100,
        }

        s2 = {
            "type": "object",
            "patternProperties": {
                r"^version$": {"type": "number"},
                r"^(?!version)\w+$": {"type": "string"},
            },
            "minProperties": 0,
            "maxProperties": 10,
        }

        expected = {
            "type": "object",
            "patternProperties": {
                r"^version$": {"type": "integer"},
                r"^(?!version)\w+$": {"type": "string"},
            },
            "minProperties": 1,
            "maxProperties": 100,
        }

        self.assertDictEqual(expected, schema_util.merge_schema(s1, s2, False))

    def test_merge_array_schema(self):
        s1 = {
            "type": "array",
            "items": [{"type": "string"}],
            "uniqueItems": True,
            "minItems": 1,
            "maxItems": 10,
        }

        s2 = {
            "type": "array",
            "items": [{"type": "number"}],
            "uniqueItems": True,
            "minItems": 10,
            "maxItems": 100,
        }

        expected = {
            "type": "array",
            "items": [{"type": "number"}],
            "uniqueItems": True,
            "minItems": 10,
            "maxItems": 100,
        }

        self.assertDictEqual(expected, schema_util.merge_schema(s1, s2))

    def test_merge_array_schema_overwrite_false(self):
        s1 = {
            "type": "array",
            "items": [{"type": "string"}],
            "uniqueItems": True,
            "minItems": 1,
            "maxItems": 10,
        }

        s2 = {
            "type": "array",
            "items": [{"type": "number"}],
            "uniqueItems": True,
            "minItems": 10,
            "maxItems": 100,
        }

        expected = {
            "type": "array",
            "items": [{"type": "string"}],
            "uniqueItems": True,
            "minItems": 1,
            "maxItems": 10,
        }

        self.assertDictEqual(expected, schema_util.merge_schema(s1, s2, False))

    def test_merge_array_schema_minimal(self):
        s1 = {"type": "array", "items": [{"type": "string"}]}

        s2 = {"type": "array", "items": [{"type": "number"}]}

        expected = {"type": "array", "items": [{"type": "number"}]}

        self.assertDictEqual(expected, schema_util.merge_schema(s1, s2))

    def test_merge_array_schema_skip_default(self):
        s1 = {
            "type": "array",
            "items": [{"type": "string"}],
            "uniqueItems": True,
            "minItems": 1,
            "maxItems": 10,
        }

        s2 = {
            "type": "array",
            "items": [{"type": "number"}],
            "uniqueItems": False,
            "minItems": 0,
            "maxItems": 0,
        }

        expected = {"type": "array", "items": [{"type": "number"}]}

        self.assertDictEqual(expected, schema_util.merge_schema(s1, s2))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.specs.native import base as test_base


class WorkflowSpecVarsValidationTest(test_base.OrchestraWorkflowSpecTest):
    def test_bad_vars_in_input(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            input:
              - y: <% ctx().a %>
            tasks:
              task1:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().a %>",
                    "message": 'Variable "a" is referenced before assignment.',
                    "schema_path": "properties.input",
                    "spec_path": "input[0]",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_seq_vars_in_input(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            input:
              - x
              - y: foobar
              - z: <% ctx().x %> <% ctx().y %>
            tasks:
              task1:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_bad_vars_with_yaql(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            input:
              - y
            vars:
              - foo: <% ctx().a %>
              - fooa: <% ctx().x + ctx().y + ctx().z %>
            tasks:
              task1:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().a %>",
                    "message": 'Variable "a" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "vars[0]",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().x + ctx().y + ctx().z %>",
                    "message": 'Variable "x" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "vars[1]",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().x + ctx().y + ctx().z %>",
                    "message": 'Variable "z" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "vars[1]",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_with_jinja(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            input:
              - y
            vars:
              - foo: "{{ ctx().a }}"
              - fooa: "{{ ctx().x + ctx().y + ctx().z }}"
            tasks:
              task1:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "jinja",
                    "expression": "{{ ctx().a }}",
                    "message": 'Variable "a" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "vars[0]",
                },
                {
                    "type": "jinja",
                    "expression": "{{ ctx().x + ctx().y + ctx().z }}",
                    "message": 'Variable "x" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "vars[1]",
                },
                {
                    "type": "jinja",
                    "expression": "{{ ctx().x + ctx().y + ctx().z }}",
                    "message": 'Variable "z" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "vars[1]",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_with_mix_yaql_and_jinja(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            input:
              - y
            vars:
              - foo: "{{ ctx().a }}"
              - fooa: <% ctx().x + ctx().y + ctx().z %>
            tasks:
              task1:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "jinja",
                    "expression": "{{ ctx().a }}",
                    "message": 'Variable "a" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "vars[0]",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().x + ctx().y + ctx().z %>",
                    "message": 'Variable "x" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "vars[1]",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().x + ctx().y + ctx().z %>",
                    "message": 'Variable "z" is referenced before assignment.',
                    "schema_path": "properties.vars",
                    "spec_path": "vars[1]",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_seq_vars_in_vars(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            vars:
              - foobar: foobar
              - fubar: <% ctx().foobar %>
              - barfoo: <% ctx().fubar %>
              - barfu: <% ctx().barfoo %>
            tasks:
              task1:
                action: core.echo
                input:
                  message: <% ctx().fubar %>
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_bad_vars_in_seq_tasks(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            vars:
              - foobar: foobar
            tasks:
              task1:
                action: core.echo
                input:
                  message: <% ctx().foobar %>
                next:
                  - when: <% succeeded() %>
                    do: task2
              task2:
                action: core.echo
                input:
                  message: <% ctx().foo %>
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo %>",
                    "message": 'Variable "foo" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task2.input",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_in_branches(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            vars:
              - foobar: foobar
            tasks:
              task1:
                action: core.echo
                input:
                  message: <% ctx().foobar %>
                next:
                  - when: <% succeeded() %>
                    publish: foo="bar"
                    do: task2
                  - when: <% failed() %>
                    publish: bar="foo"
                    do: task3
              task2:
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
              task3:
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "bar" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task2.input",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "foo" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task3.input",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_in_parallels(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            vars:
              - foobar: foobar
            tasks:
              task1:
                action: core.echo
                input:
                  message: <% ctx().foobar %>
                next:
                  - when: <% succeeded() %>
                    publish: foo="bar"
                    do: task2
              task2:
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
              task3:
                action: core.echo
                input:
                  message: <% ctx().foobar %>
                next:
                  - when: <% succeeded() %>
                    publish: bar="foo"
                    do: task4
              task4:
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "bar" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task2.input",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "foo" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task4.input",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_bad_vars_in_reusable_split_tasks(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            vars:
              - foobar: foobar
            tasks:
              task1:
                action: core.echo
                input:
                  message: <% ctx().foobar %>
                next:
                  - when: <% succeeeded() %>
                    publish: foo="bar"
                    do: task2
              task2:
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
                next:
                  - when: <% succeeded() %>
                    do: task5
              task3:
                action: core.echo
                input:
                  message: <% ctx().foobar %>
                next:
                  - when: <% succeeded() %>
                    publish: bar="foo"
                    do: task4
              task4:
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
                next:
                  - when: <% succeeded() %>
                    do: task5
              task5:
                action: core.echo
                input:
                  message: <% ctx().foo %>
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "bar" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task2.input",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo + ctx().bar %>",
                    "message": 'Variable "foo" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task4.input",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().foo %>",
                    "message": 'Variable "foo" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task5.input",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_vars_in_join_task(self):
        wf_def = """
            version: 1.0
            description: A complex branching workflow.
            vars:
              - foobar: foobar
            tasks:
              task1:
                action: core.echo
                input:
                  message: <% ctx().foobar %>
                next:
                  - when: <% succeeded() %>
                    publish: foo="bar"
                    do: task2
              task2:
                action: core.echo
                input:
                  message: <% ctx().foo %>
                next:
                  - when: <% succeeded() %>
                    do: task5
              task3:
                action: core.echo
                input:
                  message: <% ctx().foobar %>
                next:
                  - when: <% succeeded() %>
                    publish: bar="foo"
                    do: task4
              task4:
                action: core.echo
                input:
                  message: <% ctx().bar %>
                next:
                  - when: <% succeeded() %>
                    do: task5
              task5:
                join: all
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), dict())

    def test_bad_vars_in_publish(self):
        wf_def = """
            version: 1.0
            description: A basic workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - publish: foo="bar" bar="foo"
                    do: task2
              task2:
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
                next:
                  - publish: foobar=<% ctx().fu + ctx().bar %>
                    do: task3
              task3:
                action: core.echo
                input:
                  message: <% ctx().fu + ctx().bar %>
                next:
                  - publish: foobar=<% ctx().fu + ctx().bar %>
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().fu + ctx().bar %>",
                    "message": 'Variable "fu" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task3.input",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().fu + ctx().bar %>",
                    "message": 'Variable "fu" is referenced before assignment.',
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$.properties."
                        "next.items.properties.publish"
                    ),
                    "spec_path": "tasks.task2.next[0].publish[0]",
                },
                {
                    "type": "yaql",
                    "expression": "<% ctx().fu + ctx().bar %>",
                    "message": 'Variable "fu" is referenced before assignment.',
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$.properties."
                        "next.items.properties.publish"
                    ),
                    "spec_path": "tasks.task3.next[0].publish[0]",
                },
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_vars_in_publish_inline_params(self):
        wf_def = """
            version: 1.0
            description: A basic workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - publish: foo="bar" bar="foo"
                    do: task2, task3
              task2:
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
              task3:
                action: core.echo
                input:
                  message: <% ctx().fu + ctx().bar %>
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().fu + ctx().bar %>",
                    "message": 'Variable "fu" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.input",
                    "spec_path": "tasks.task3.input",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_seq_vars_in_publish(self):
        wf_def = """
            version: 1.0
            description: A basic workflow.
            tasks:
              task1:
                action: core.noop
                next:
                  - publish: foo="bar" bar="foo" foobar="<% ctx().foo %><% ctx().bar %>"
                    do: task2
              task2:
                action: core.echo
                input:
                  message: <% ctx().foo + ctx().bar %>
                next:
                  - publish:
                      - fu: "fu"
                      - fubar: "<% ctx().fu %><% ctx().bar %>"
                    do: task3
              task3:
                action: core.echo
                input:
                  message: <% ctx().fu %><% ctx().bar %>
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_bad_vars_in_output(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            output:
              - y: <% ctx().a %>
            tasks:
              task1:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().a %>",
                    "message": 'Variable "a" is referenced before assignment.',
                    "schema_path": "properties.output",
                    "spec_path": "output[0]",
                }
            ]
        }

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_seq_vars_in_output(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            input:
              - a
            output:
              - x: <% ctx().a %>
              - y: <% ctx().x %>
              - z: <% ctx().x %> <% ctx().y %>
            tasks:
              task1:
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

    def test_vars_in_output_on_error_handling(self):
        wf_def = """
            version: 1.0
            description: A basic sequential workflow.
            input:
              - a
            output:
              - b: <% ctx().b %>
              - x: <% ctx().a %>
              - y: <% ctx().x %>
              - z: <% ctx().x %> <% ctx().y %>
            tasks:
              task1:
                action: core.noop
                next:
                  - when: <% failed() %>
                    publish: b="foobar"
                    do: fail
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})



# Copyright 2019 The Linux Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.specs.mistral.v2 import policies as policy_models
from orquesta.specs.mistral.v2 import tasks as task_models
from orquesta.tests.unit.specs.mistral import base as test_base


class TaskSpecTest(test_base.MistralWorkflowSpecTest):
    def test_simplest_task(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: The simplest workflow
                tasks:
                    task1:
                        action: std.noop
        """

        task_spec = self.instantiate(wf_def).tasks.get_task("task1")

        self.assertIsInstance(task_spec, task_models.TaskSpec)
        self.assertIsNone(task_spec.has_retry())

    def test_retry_task(self):
        wf_def = """
            version: '2.0'

            sequential:
                description: Workflow which has a retrying task
                tasks:
                    task1:
                        action: std.noop
                        retry:
                            delay: 5
                            count: 10
        """

        task_spec = self.instantiate(wf_def).tasks.get_task("task1")
        task_retry = task_spec.has_retry()

        self.assertIsInstance(task_retry, policy_models.RetrySpec)
        self.assertEqual(task_retry.delay, 5)
        self.assertEqual(task_retry.count, 10)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit import base as test_base


class JinjaFacadeValidationTest(test_base.ExpressionFacadeEvaluatorTest):
    def test_basic_validate(self):
        self.assertListEqual([], self.validate(None))
        self.assertListEqual([], self.validate("{{ 1 }}"))
        self.assertListEqual([], self.validate("{{ abc }}"))
        self.assertListEqual([], self.validate("{{ 1 + 2 }}"))
        self.assertListEqual([], self.validate("{{ ctx().foo }}"))
        self.assertListEqual([], self.validate('{{ ctx("foo") }}'))
        self.assertListEqual([], self.validate('{{ ctx().a1 + ctx("a2") }}'))

    def test_parse_error(self):
        expr = "{{ {{ ctx().foo }} }}"
        errors = self.validate(expr)

        self.assertEqual(2, len(errors))
        self.assertIn("expected token ':', got '}'", errors[0]["message"])
        self.assertIn("unexpected end of template", errors[1]["message"])

    def test_multiple_errors(self):
        expr = "{{ 1 +/ 2 }} and {{ * }}"
        errors = self.validate(expr)

        self.assertEqual(3, len(errors))
        self.assertIn("unexpected", errors[0]["message"])
        self.assertIn("unexpected", errors[1]["message"])
        self.assertIn("unexpected", errors[2]["message"])

    def test_block_error(self):
        expr = "{% for i in ctx().x %}{{ i }}{% foobar %}"
        errors = self.validate(expr)

        self.assertEqual(1, len(errors))
        self.assertIn("unknown tag", errors[0]["message"])

    def test_missing_braces_error(self):
        expr = "{% for i in ctx().x %}{{ i }}{{ foobar %}"
        errors = self.validate(expr)

        self.assertEqual(1, len(errors))
        self.assertIn("unexpected '}'", errors[0]["message"])

    def test_list_validate(self):
        self.assertListEqual([], self.validate([]))

        target = ["{{ 1 }}", "xyz", 123, True]
        self.assertListEqual([], self.validate(target))

        target = ["{{ 1 }}", ["{{ a }}", "x"]]
        self.assertListEqual([], self.validate(target))

        target = ["{{ 1 }}", ["{{ a }}", "x"], {"k1": "{{ 3 }}"}]
        self.assertListEqual([], self.validate(target))

    def test_list_multiple_errors(self):
        target = ["{{ 1 +/ 2 }}", ["{{ {{ ctx().foo }}"], {"k1": "{{ * }}"}]
        result = self.validate(target)
        self.assertEqual(3, len(result))

    def test_dict_validate(self):
        self.assertListEqual([], self.validate({}))

        target = {
            "k1": "{{ 1 }}",
            "k2": "foobar",
            "{{ k3 }}": 789,
            "k4": ["{{ abc }}", "xyz", 123, False, {"k": "v"}],
            "depth-1": {
                "depth-1-1": {
                    "depth-1-1-2": '{{ ctx().a1 + ctx("a2") }}',
                    "depth-1-1-3": ["{{ ctx().foobar }}", "xyz"],
                }
            },
        }

        self.assertListEqual([], self.validate(target))

    def test_dict_multiple_errors(self):
        target = {
            "k1": "{{ 1 +/ 2 }}",
            "k2": "foobar",
            "{{ {{ ctx().foo }}": 789,
            "k4": ["{{ abc }}", "xyz", 123, False, {"k": "v"}],
            "depth-1": {
                "depth-1-1": {
                    "depth-1-1-2": "{{ 3 +/ 4 }}",
                    "depth-1-1-3": ["{{ ctx().foobar }}", "{{ * }}"],
                }
            },
        }

        result = self.validate(target)
        self.assertEqual(4, len(result))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import itertools
import logging
import re
import six

import yaql
import yaql.language.exceptions as yaql_exc
import yaql.language.utils as yaql_utils

from orquesta import exceptions as exc
from orquesta.expressions import base as expr_base
from orquesta.expressions.functions import base as func_base
from orquesta.utils import expression as expr_util
from orquesta.utils import strings as str_util


LOG = logging.getLogger(__name__)


def register_functions(ctx):
    catalog = func_base.load()

    for name, func in six.iteritems(catalog):
        ctx.register_function(func, name=name)

    return catalog


class YaqlGrammarException(exc.ExpressionGrammarException):
    pass


class YaqlEvaluationException(exc.ExpressionEvaluationException):
    pass


class YAQLEvaluator(expr_base.Evaluator):
    _type = "yaql"
    _delimiter = "<%>"
    _regex_pattern = "<%.*?%>"
    _regex_parser = re.compile(_regex_pattern)

    _regex_ctx_ref_pattern = r'[][a-zA-Z0-9_\'"\.()]*'
    # match any of:
    #   word boundary ctx(*)
    #   word boundary ctx()*
    #   word boundary ctx().*
    #   word boundary ctx(*)*
    #   word boundary ctx(*).*
    _regex_ctx_pattern = r'\bctx\([\'"]?{0}[\'"]?\)\.?{0}'.format(_regex_ctx_ref_pattern)
    _regex_ctx_var_parser = re.compile(_regex_ctx_pattern)

    _regex_var = r"[a-zA-Z0-9_-]+"
    _regex_var_extracts = [
        r"(?<=\bctx\(\)\.)({})\b(?!\()\.?".format(_regex_var),  # extract x in ctx().x
        r"(?:\bctx\(({})\))".format(_regex_var),  # extract x in ctx(x)
        r"(?:\bctx\(\'({})\'\))".format(_regex_var),  # extract x in ctx('x')
        r'(?:\bctx\("({})"\))'.format(_regex_var),  # extract x in ctx("x")
    ]

    _engine = yaql.language.factory.YaqlFactory().create()
    _root_ctx = yaql.create_context()
    _custom_functions = register_functions(_root_ctx)

    @classmethod
    def contextualize(cls, data):
        ctx = cls._root_ctx.create_child_context()

        # Some yaql expressions (e.g. distinct()) refer to hash value of variable.
        # But some built-in Python type values (e.g. list and dict) don't have __hash__() method.
        # The convert_input_data method parses specified variable and convert it to hashable one.
        if isinstance(data, yaql_utils.SequenceType) or isinstance(data, yaql_utils.MappingType):
            ctx["__vars"] = yaql_utils.convert_input_data(data)
        else:
            ctx["__vars"] = data or {}

        ctx["__state"] = ctx["__vars"].get("__state")
        ctx["__current_task"] = ctx["__vars"].get("__current_task")
        ctx["__current_item"] = ctx["__vars"].get("__current_item")

        return ctx

    @classmethod
    def get_statement_regex(cls):
        return cls._regex_pattern

    @classmethod
    def has_expressions(cls, text):
        exprs = cls._regex_parser.findall(text)

        return exprs is not None and len(exprs) > 0

    @classmethod
    def get_var_extraction_regexes(cls):
        return cls._regex_var_extracts

    @classmethod
    def validate(cls, text):
        if not isinstance(text, six.string_types):
            raise ValueError("Text to be evaluated is not typeof string.")

        errors = []

        for expr in cls._regex_parser.findall(text):
            try:
                cls._engine(cls.strip_delimiter(expr))
            except (yaql_exc.YaqlException, ValueError, TypeError) as e:
                errors.append(expr_util.format_error(cls._type, expr, e))

        return errors

    @classmethod
    def evaluate(cls, text, data=None):
        if not isinstance(text, six.string_types):
            raise ValueError("Text to be evaluated is not typeof string.")

        if data and not isinstance(data, dict):
            raise ValueError("Provided data is not typeof dict.")

        output = str_util.unicode(text)
        exprs = cls._regex_parser.findall(text)
        ctx = cls.contextualize(data)

        try:
            for expr in exprs:
                stripped = cls.strip_delimiter(expr)
                result = cls._engine(stripped).evaluate(context=ctx)

                if inspect.isgenerator(result):
                    result = list(result)

                if isinstance(result, six.string_types):
                    result = cls.evaluate(result, data)

                if len(exprs) > 1 or len(output) > len(expr):
                    output = output.replace(expr, str_util.unicode(result, force=True))
                else:
                    output = str_util.unicode(result)

        except KeyError as e:
            error = str(getattr(e, "message", e)).strip("'")
            msg = "Unable to resolve key '%s' in expression '%s' from context."
            raise YaqlEvaluationException(msg % (error, expr))
        except (yaql_exc.YaqlException, ValueError, TypeError) as e:
            msg = "Unable to evaluate expression '%s'. %s: %s"
            raise YaqlEvaluationException(msg % (expr, e.__class__.__name__, str(e).strip("'")))
        except Exception as e:
            msg = "Unable to evaluate expression '%s'. %s: %s"
            raise YaqlEvaluationException(msg % (expr, e.__class__.__name__, str(e)))

        return output

    @classmethod
    def extract_vars(cls, text):
        if not isinstance(text, six.string_types):
            raise ValueError("Text to be evaluated is not typeof string.")

        results = [
            cls._regex_ctx_var_parser.findall(expr.strip(cls._delimiter).strip())
            for expr in cls._regex_parser.findall(text)
        ]

        variables = [v.strip() for v in itertools.chain.from_iterable(results)]

        return sorted(list(set(variables)))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta import constants
from orquesta import exceptions as exc
from orquesta.expressions.functions import workflow as funcs
from orquesta import statuses
from orquesta.utils import jsonify as json_util


class WorkflowFunctionTest(unittest.TestCase):
    def test_task_status_empty_context(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))

        context = None
        self.assertEqual(funcs.task_status_(context, task_name, task_route), statuses.UNSET)

        context = {}
        self.assertEqual(funcs.task_status_(context, task_name, task_route), statuses.UNSET)

        context = {"__state": None}
        self.assertEqual(funcs.task_status_(context, task_name, task_route), statuses.UNSET)

        context = {"__state": {}}
        self.assertEqual(funcs.task_status_(context, task_name, task_route), statuses.UNSET)

        context = {"__state": {"tasks": None}}
        self.assertEqual(funcs.task_status_(context, task_name, task_route), statuses.UNSET)

        context = {"__state": {"tasks": {}}}
        self.assertEqual(funcs.task_status_(context, task_name, task_route), statuses.UNSET)

        task_pointers = {task_flow_pointer_id: None}
        context = {"__state": {"tasks": task_pointers}}
        self.assertEqual(funcs.task_status_(context, task_name, task_route), statuses.UNSET)

        task_pointers = {task_flow_pointer_id: 0}
        context = {"__state": {"tasks": task_pointers, "sequence": [None]}}
        self.assertEqual(funcs.task_status_(context, task_name, task_route), statuses.UNSET)

    def test_task_status_empty_context_with_no_given_route(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))
        task_pointers = {task_flow_pointer_id: None}

        context_test_cases = [
            None,
            {},
            {"__state": None},
            {"__state": {}},
            {"__state": {"tasks": None}},
            {"__state": {"tasks": task_pointers}},
            {"__state": {"tasks": task_pointers, "sequence": [None]}},
        ]

        for context in context_test_cases:
            self.assertEqual(funcs.task_status_(context, task_name), statuses.UNSET)

    def test_task_status_empty_context_with_no_given_route_but_current_task_set(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))
        task_pointers = {task_flow_pointer_id: None}

        context_test_cases = [
            {},
            {"__state": None},
            {"__state": {}},
            {"__state": {"tasks": None}},
            {"__state": {"tasks": task_pointers}},
            {"__state": {"tasks": task_pointers, "sequence": [None]}},
        ]

        for context in context_test_cases:
            context["__current_task"] = {"id": task_name, "route": task_route}
            self.assertEqual(funcs.task_status_(context, task_name), statuses.UNSET)

    def test_task_status_dereference_errors(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))

        task_pointers = {task_flow_pointer_id: 0}
        context = {"__state": {"tasks": task_pointers}}
        self.assertRaises(IndexError, funcs.task_status_, context, task_name, task_route)

        task_pointers = {task_flow_pointer_id: 1}
        context = {"__state": {"tasks": task_pointers, "sequence": [{"status": statuses.RUNNING}]}}
        self.assertRaises(IndexError, funcs.task_status_, context, task_name, task_route)

    def test_task_status(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))

        task_pointers = {task_flow_pointer_id: 0}
        context = {"__state": {"tasks": task_pointers, "sequence": [{"status": statuses.RUNNING}]}}
        actual_task_status = funcs.task_status_(context, task_name, route=task_route)
        self.assertEqual(actual_task_status, statuses.RUNNING)

    def test_task_status_route_from_current_task(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))

        task_pointers = {task_flow_pointer_id: 0}
        context = {
            "__current_task": {"id": task_name, "route": task_route},
            "__state": {"tasks": task_pointers, "sequence": [{"status": statuses.RUNNING}]},
        }

        actual_task_status = funcs.task_status_(context, task_name)
        self.assertEqual(actual_task_status, statuses.RUNNING)

    def test_get_current_task_empty(self):
        self.assertRaises(exc.ExpressionEvaluationException, funcs._get_current_task, None)

        context = {"__current_task": None}
        self.assertRaises(exc.ExpressionEvaluationException, funcs._get_current_task, context)

        context = {"__current_task": {}}
        self.assertRaises(exc.ExpressionEvaluationException, funcs._get_current_task, context)

    def test_get_current_task(self):
        context = {"__current_task": {"id": "t1", "route": 0}}
        self.assertDictEqual(funcs._get_current_task(context), context["__current_task"])

    def test_succeeded(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))

        context = {"__current_task": None}
        self.assertRaises(exc.ExpressionEvaluationException, funcs.succeeded_, context)

        context = {
            "__current_task": {"id": task_name, "route": task_route},
            "__state": {
                "tasks": {task_flow_pointer_id: 0},
                "sequence": [{"status": statuses.RUNNING}],
            },
        }

        self.assertFalse(funcs.succeeded_(context))

        context = {
            "__current_task": {"id": task_name, "route": task_route},
            "__state": {
                "tasks": {task_flow_pointer_id: 0},
                "sequence": [{"status": statuses.SUCCEEDED}],
            },
        }

        self.assertTrue(funcs.succeeded_(context))

    def test_failed(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))

        context = {"__current_task": None}
        self.assertRaises(exc.ExpressionEvaluationException, funcs.failed_, context)

        context = {
            "__current_task": {"id": task_name, "route": task_route},
            "__state": {
                "tasks": {task_flow_pointer_id: 0},
                "sequence": [{"status": statuses.RUNNING}],
            },
        }

        self.assertFalse(funcs.failed_(context))

        context = {
            "__current_task": {"id": task_name, "route": task_route},
            "__state": {
                "tasks": {task_flow_pointer_id: 0},
                "sequence": [{"status": statuses.FAILED}],
            },
        }

        self.assertTrue(funcs.failed_(context))

    def test_completed(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))

        context = {"__current_task": None}
        self.assertRaises(exc.ExpressionEvaluationException, funcs.completed_, context)

        context = {
            "__current_task": {"id": task_name, "route": task_route},
            "__state": {
                "tasks": {task_flow_pointer_id: 0},
                "sequence": [{"status": statuses.RUNNING}],
            },
        }

        self.assertFalse(funcs.completed_(context))

        context = {
            "__current_task": {"id": task_name, "route": task_route},
            "__state": {
                "tasks": {task_flow_pointer_id: 0},
                "sequence": [{"status": statuses.SUCCEEDED}],
            },
        }

        self.assertTrue(funcs.completed_(context))

    def test_result(self):
        task_route = 0
        task_name = "t1"
        task_flow_pointer_id = constants.TASK_STATE_ROUTE_FORMAT % (task_name, str(task_route))

        context = {"__current_task": None}
        self.assertRaises(exc.ExpressionEvaluationException, funcs.result_, context)

        context = {
            "__current_task": {"id": task_name, "route": task_route, "result": "foobar"},
            "__state": {
                "tasks": {task_flow_pointer_id: 0},
                "sequence": [{"status": statuses.SUCCEEDED}],
            },
        }

        self.assertEqual(funcs.result_(context), "foobar")

        context = {
            "__current_task": {"id": task_name, "route": task_route, "result": {"fu": "bar"}},
            "__state": {
                "tasks": {task_flow_pointer_id: 0},
                "sequence": [{"status": statuses.SUCCEEDED}],
            },
        }

        self.assertDictEqual(funcs.result_(context), {"fu": "bar"})

    def test_task_status_of_tasks_along_split(self):
        task_pointers = {
            "t1__r0": 0,
            "t2__r0": 1,
            "t3__r0": 2,
            "t4__r1": 3,
            "t4__r2": 4,
            "t5__r1": 5,
            "t5__r2": 6,
            "t6__r1": 7,
            "t6__r2": 8,
        }

        task_flow_entries = [
            {"id": "t1", "route": 0, "status": statuses.SUCCEEDED},
            {"id": "t2", "route": 0, "status": statuses.SUCCEEDED},
            {"id": "t3", "route": 0, "status": statuses.SUCCEEDED},
            {"id": "t4", "route": 1, "status": statuses.SUCCEEDED},
            {"id": "t4", "route": 2, "status": statuses.FAILED},
            {"id": "t5", "route": 1, "status": statuses.SUCCEEDED},
            {"id": "t5", "route": 2, "status": statuses.FAILED},
            {"id": "t6", "route": 1, "status": statuses.SUCCEEDED},
            {"id": "t6", "route": 2, "status": statuses.FAILED},
        ]

        routes = [
            [],
            ["t2__t0"],
            ["t3__t0"],
        ]

        context = {
            "__state": {"tasks": task_pointers, "sequence": task_flow_entries, "routes": routes}
        }

        # Check the task statuses along route 1.
        current_ctx = json_util.deepcopy(context)
        current_ctx["__current_task"] = {"id": "t6", "route": 1}
        self.assertEqual(funcs.task_status_(current_ctx, "t1"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t2"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t3"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t4"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t5"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t6"), statuses.SUCCEEDED)

        # Check the task statuses along route 2.
        current_ctx = json_util.deepcopy(context)
        current_ctx["__current_task"] = {"id": "t6", "route": 2}
        self.assertEqual(funcs.task_status_(current_ctx, "t1"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t2"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t3"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t4"), statuses.FAILED)
        self.assertEqual(funcs.task_status_(current_ctx, "t5"), statuses.FAILED)
        self.assertEqual(funcs.task_status_(current_ctx, "t6"), statuses.FAILED)

    def test_task_status_of_tasks_along_splits(self):
        task_pointers = {
            "t1__r0": 0,
            "t2__r0": 1,
            "t3__r0": 2,
            "t4__r1": 3,
            "t4__r2": 4,
            "t5__r1": 5,
            "t5__r2": 6,
            "t6__r1": 7,
            "t6__r2": 8,
            "t7__r3": 9,
            "t7__r4": 10,
            "t7__r5": 11,
            "t7__r6": 12,
        }

        task_flow_entries = [
            {"id": "t1", "route": 0, "status": statuses.SUCCEEDED},
            {"id": "t2", "route": 0, "status": statuses.SUCCEEDED},
            {"id": "t3", "route": 0, "status": statuses.SUCCEEDED},
            {"id": "t4", "route": 1, "status": statuses.SUCCEEDED},
            {"id": "t4", "route": 2, "status": statuses.FAILED},
            {"id": "t5", "route": 1, "status": statuses.SUCCEEDED},
            {"id": "t5", "route": 2, "status": statuses.FAILED},
            {"id": "t6", "route": 1, "status": statuses.SUCCEEDED},
            {"id": "t6", "route": 2, "status": statuses.FAILED},
            {"id": "t7", "route": 3, "status": statuses.SUCCEEDED},
            {"id": "t7", "route": 4, "status": statuses.FAILED},
            {"id": "t7", "route": 5, "status": statuses.RUNNING},
            {"id": "t7", "route": 6, "status": statuses.DELAYED},
        ]

        routes = [
            [],
            ["t2__t0"],
            ["t3__t0"],
            ["t2__t0", "t5__t0"],
            ["t3__t0", "t5__t0"],
            ["t2__t0", "t6__t0"],
            ["t3__t0", "t6__t0"],
        ]

        context = {
            "__state": {"tasks": task_pointers, "sequence": task_flow_entries, "routes": routes}
        }

        # Check the task statuses along route 1.
        current_ctx = json_util.deepcopy(context)
        current_ctx["__current_task"] = {"id": "t7", "route": 3}
        self.assertEqual(funcs.task_status_(current_ctx, "t1"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t2"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t3"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t4"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t5"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t6"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t7"), statuses.SUCCEEDED)

        # Check the task statuses along route 2.
        current_ctx = json_util.deepcopy(context)
        current_ctx["__current_task"] = {"id": "t7", "route": 4}
        self.assertEqual(funcs.task_status_(current_ctx, "t1"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t2"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t3"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t4"), statuses.FAILED)
        self.assertEqual(funcs.task_status_(current_ctx, "t5"), statuses.FAILED)
        self.assertEqual(funcs.task_status_(current_ctx, "t6"), statuses.FAILED)
        self.assertEqual(funcs.task_status_(current_ctx, "t7"), statuses.FAILED)

        # Check the task statuses along route 3.
        current_ctx = json_util.deepcopy(context)
        current_ctx["__current_task"] = {"id": "t7", "route": 5}
        self.assertEqual(funcs.task_status_(current_ctx, "t1"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t2"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t3"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t4"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t5"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t6"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t7"), statuses.RUNNING)

        # Check the task statuses along route 4.
        current_ctx = json_util.deepcopy(context)
        current_ctx["__current_task"] = {"id": "t7", "route": 6}
        self.assertEqual(funcs.task_status_(current_ctx, "t1"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t2"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t3"), statuses.SUCCEEDED)
        self.assertEqual(funcs.task_status_(current_ctx, "t4"), statuses.FAILED)
        self.assertEqual(funcs.task_status_(current_ctx, "t5"), statuses.FAILED)
        self.assertEqual(funcs.task_status_(current_ctx, "t6"), statuses.FAILED)
        self.assertEqual(funcs.task_status_(current_ctx, "t7"), statuses.DELAYED)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.specs import mistral as mistral_specs
from orquesta.tests.unit.specs.mistral import base as test_base


class WorkflowSpecTest(test_base.MistralWorkflowSpecTest):
    def test_exception_empty_definition(self):
        self.assertRaises(ValueError, mistral_specs.WorkflowSpec, None, {})
        self.assertRaises(ValueError, mistral_specs.WorkflowSpec, None, "")
        self.assertRaises(ValueError, mistral_specs.WorkflowSpec, None, None)

    def test_basic_spec_serialization(self):
        wf_name = "sequential"
        wf_spec_1 = self.get_wf_spec(wf_name)
        wf_spec_2 = mistral_specs.WorkflowSpec.deserialize(wf_spec_1.serialize())

        self.assertDictEqual(wf_spec_2.serialize(), wf_spec_1.serialize())

    def test_get_next_tasks(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertListEqual(
            wf_spec.tasks.get_next_tasks("task1"),
            [("task2", None, "on-success"), ("task3", None, "on-success")],
        )

        self.assertListEqual(wf_spec.tasks.get_next_tasks("task2"), [("task4", None, "on-success")])

        self.assertListEqual(wf_spec.tasks.get_next_tasks("task3"), [("task4", None, "on-success")])

        self.assertListEqual(
            wf_spec.tasks.get_next_tasks("task4"),
            [("task5", None, "on-success"), ("task6", None, "on-success")],
        )

        self.assertListEqual(wf_spec.tasks.get_next_tasks("task5"), [("task7", None, "on-success")])

        self.assertListEqual(wf_spec.tasks.get_next_tasks("task6"), [("task7", None, "on-success")])

        self.assertListEqual(wf_spec.tasks.get_next_tasks("task7"), [])

    def test_get_prev_tasks(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertListEqual(wf_spec.tasks.get_prev_tasks("task1"), [])

        self.assertListEqual(wf_spec.tasks.get_prev_tasks("task2"), [("task1", None, "on-success")])

        self.assertListEqual(wf_spec.tasks.get_prev_tasks("task3"), [("task1", None, "on-success")])

        self.assertListEqual(
            wf_spec.tasks.get_prev_tasks("task4"),
            [("task2", None, "on-success"), ("task3", None, "on-success")],
        )

        self.assertListEqual(wf_spec.tasks.get_prev_tasks("task5"), [("task4", None, "on-success")])

        self.assertListEqual(wf_spec.tasks.get_prev_tasks("task6"), [("task4", None, "on-success")])

        self.assertListEqual(
            wf_spec.tasks.get_prev_tasks("task7"),
            [("task5", None, "on-success"), ("task6", None, "on-success")],
        )

    def test_get_start_tasks(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertListEqual(wf_spec.tasks.get_start_tasks(), [("task1", None, None)])

    def test_is_join_task(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertFalse(wf_spec.tasks.is_join_task("task4"))
        self.assertTrue(wf_spec.tasks.is_join_task("task7"))

    def test_is_split_task(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertTrue(wf_spec.tasks.is_split_task("task4"))
        self.assertFalse(wf_spec.tasks.is_split_task("task7"))

    def test_not_in_cycle(self):
        wf_name = "split"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertFalse(wf_spec.tasks.in_cycle("task4"))

    def test_has_cycles(self):
        wf_name = "cycle"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertTrue(wf_spec.tasks.has_cycles())

    def test_in_cycle(self):
        wf_name = "cycle"
        wf_spec = self.get_wf_spec(wf_name)

        self.assertTrue(wf_spec.tasks.in_cycle("task1"))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from six.moves import queue

from orquesta.composers import base as comp_base
from orquesta import graphing
from orquesta.specs import mistral as mistral_specs
from orquesta import statuses


LOG = logging.getLogger(__name__)

TASK_TRANSITION_MAP = {
    "on-success": [statuses.SUCCEEDED],
    "on-error": statuses.ABENDED_STATUSES,
    "on-complete": statuses.COMPLETED_STATUSES,
}


class WorkflowComposer(comp_base.WorkflowComposer):
    wf_spec_type = mistral_specs.WorkflowSpec

    @classmethod
    def compose(cls, spec):
        if not cls.wf_spec_type:
            raise TypeError("Undefined spec type for composer.")

        if not isinstance(spec, cls.wf_spec_type):
            raise TypeError('Unsupported spec type "%s".' % str(type(spec)))

        return cls._compose_wf_graph(spec)

    @classmethod
    def _compose_transition_criteria(cls, task_name, *args, **kwargs):
        criteria = []

        condition = kwargs.get("condition")
        expr = kwargs.get("expr")

        task_status_criterion = "task_status(%s) in %s" % (
            task_name,
            str(TASK_TRANSITION_MAP[condition]),
        )

        criteria.append("<% " + task_status_criterion + " %>")

        if expr:
            criteria.append(expr)

        return criteria

    @classmethod
    def _compose_wf_graph(cls, wf_spec):
        if not isinstance(wf_spec, cls.wf_spec_type):
            raise TypeError("Workflow spec is not typeof %s." % cls.wf_spec_type.__name__)

        q = queue.Queue()
        wf_graph = graphing.WorkflowGraph()

        for task_name, expr, condition in wf_spec.tasks.get_start_tasks():
            q.put((task_name, []))

        while not q.empty():
            task_name, splits = q.get()

            wf_graph.add_task(task_name)

            if wf_spec.tasks.is_join_task(task_name):
                task_spec = wf_spec.tasks[task_name]
                barrier = "*" if task_spec.join == "all" else task_spec.join
                wf_graph.set_barrier(task_name, value=barrier)

            # Determine if the task is a split task and if it is in a cycle. If the task is a
            # split task, keep track of where the split(s) occurs.
            if wf_spec.tasks.is_split_task(task_name) and not wf_spec.tasks.in_cycle(task_name):
                splits.append(task_name)

            if splits:
                wf_graph.update_task(task_name, splits=splits)

            next_tasks = wf_spec.tasks.get_next_tasks(task_name)

            for next_task_name, expr, condition in next_tasks:
                if not wf_graph.has_task(next_task_name) or not wf_spec.tasks.in_cycle(
                    next_task_name
                ):
                    q.put((next_task_name, list(splits)))

                crta = cls._compose_transition_criteria(task_name, condition=condition, expr=expr)
                seqs = wf_graph.has_transition(task_name, next_task_name, criteria=crta)

                # Use existing transition if present otherwise create new transition.
                if seqs:
                    wf_graph.update_transition(
                        task_name, next_task_name, key=seqs[0][2], criteria=crta
                    )
                else:
                    wf_graph.add_transition(task_name, next_task_name, criteria=crta)

        return wf_graph



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import constants


class TaskRerunRequest(object):
    def __init__(self, task_id, route, reset_items=False):
        self.task_id = task_id
        self.route = route
        self.reset_items = reset_items
        self.task_state_entry_id = constants.TASK_STATE_ROUTE_FORMAT % (task_id, str(route))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.expressions import yql as yaql_expr
from orquesta.tests.unit import base as test_base
from orquesta.utils import plugin as plugin_util


class YAQLEvaluationTest(test_base.ExpressionEvaluatorTest):
    @classmethod
    def setUpClass(cls):
        cls.language = "yaql"
        super(YAQLEvaluationTest, cls).setUpClass()

    def test_get_evaluator(self):
        e = plugin_util.get_module("orquesta.expressions.evaluators", self.language)

        self.assertEqual(e, yaql_expr.YAQLEvaluator)
        self.assertIn("json", e._custom_functions.keys())

    def test_custom_function(self):
        expr = "<% json('{\"a\": 123}') %>"

        self.assertDictEqual({"a": 123}, self.evaluator.evaluate(expr))

    def test_custom_function_failure(self):
        expr = "<% json(int(123)) %>"

        self.assertRaises(yaql_expr.YaqlEvaluationException, self.evaluator.evaluate, expr)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.conducting.mistral import base


class JinjaCriteriaConductorTest(base.MistralWorkflowConductorTest):
    def test_decision_tree_with_jinja_expr(self):
        wf_name = "decision-jinja"

        # Test branch "a"
        expected_task_seq = ["t1", "a"]

        self.assert_conducting_sequences(wf_name, expected_task_seq, inputs={"which": "a"})

        # Test branch "b"
        expected_task_seq = ["t1", "b"]

        self.assert_conducting_sequences(wf_name, expected_task_seq, inputs={"which": "b"})

        # Test branch "c"
        expected_task_seq = ["t1", "c"]

        self.assert_conducting_sequences(wf_name, expected_task_seq, inputs={"which": "c"})



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import statuses
from orquesta.tests.unit.conducting.native import base


class JoinWorkflowConductorTest(base.OrchestraWorkflowConductorTest):
    def test_join(self):
        wf_name = "join"

        self.assert_spec_inspection(wf_name)

        # Mock successful join.
        expected_task_seq = ["task1", "task2", "task4", "task3", "task5", "task6", "task7"]

        self.assert_conducting_sequences(wf_name, expected_task_seq)

        # Both tasks before the join, task3 and task5, failed.
        # The criteria for the join task is not met.
        expected_task_seq = ["task1", "task2", "task4", "task3", "task5"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task3
            statuses.FAILED,  # task5
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.FAILED,
        )

        # First task before the join, task3, failed.
        # The criteria for the join task is not met.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task3
            statuses.FAILED,  # task5
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.FAILED,
        )

        # Second task before the join, task5, failed.
        # The criteria for the join task is not met.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task3
            statuses.SUCCEEDED,  # task5
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.FAILED,
        )

    def test_join_count(self):
        wf_name = "join-count"

        self.assert_spec_inspection(wf_name)

        # Mock one of the branches is still running and
        # the join task is satisfied and completed.
        expected_task_seq = ["task1", "task2", "task4", "task6", "task3", "task5", "task8"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.RUNNING,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.SUCCEEDED,  # task8
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.RUNNING,
        )

        # Mock error at task6 of branch 3. The criteria for join task is met.
        expected_task_seq = [
            "task1",
            "task2",
            "task4",
            "task6",
            ("noop", 1),
            "task3",
            "task5",
            "task8",
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.SUCCEEDED,  # task8
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task6__t0"]],
        )

        # Mock error at task7 of branch 3. The criteria for join task is not met.
        expected_task_seq = [
            "task1",
            "task2",
            "task4",
            "task6",
            "task3",
            "task5",
            "task7",
            ("noop", 1),
            "task8",
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.FAILED,  # task7
            statuses.SUCCEEDED,  # task8
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task7__t0"]],
        )

        # Mock errors at branch 2 and 3. The criteria for join task is not met.
        expected_task_seq = [
            "task1",
            "task2",
            "task4",
            "task6",
            "task3",
            "task5",
            ("noop", 1),
            "task7",
            ("noop", 2),
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.FAILED,  # task5
            statuses.FAILED,  # task7
        ]

        conductor = self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task5__t0"], ["task7__t0"]],
            expected_workflow_status=statuses.FAILED,
        )

        expected_errors = [
            {
                "message": "Execution failed. See result for details.",
                "type": "error",
                "task_id": "task5",
            },
            {
                "message": "Execution failed. See result for details.",
                "type": "error",
                "task_id": "task7",
            },
            {
                "message": (
                    'UnreachableJoinError: The join task|route "task8|0" '
                    "is partially satisfied but unreachable."
                ),
                "type": "error",
                "route": 0,
                "task_id": "task8",
            },
        ]

        self.assertListEqual(conductor.errors, expected_errors)

        # Mock all branches failed. The criteria for join task is not met.
        expected_task_seq = [
            "task1",
            "task2",
            "task4",
            "task6",
            "task3",
            "task5",
            ("noop", 1),
            "task7",
            ("noop", 2),
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task6
            statuses.FAILED,  # task3
            statuses.FAILED,  # task5
            statuses.FAILED,  # task7
        ]

        conductor = self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task5__t0"], ["task7__t0"]],
            expected_workflow_status=statuses.FAILED,
        )

        expected_errors = [
            {
                "message": "Execution failed. See result for details.",
                "type": "error",
                "task_id": "task3",
            },
            {
                "message": "Execution failed. See result for details.",
                "type": "error",
                "task_id": "task5",
            },
            {
                "message": "Execution failed. See result for details.",
                "type": "error",
                "task_id": "task7",
            },
        ]

        self.assertListEqual(conductor.errors, expected_errors)

    def test_join_context(self):
        wf_name = "join-context"

        expected_output = {
            "messages": [
                "Fee fi fo fum",
                "I smell the blood of an English man",
                "Be alive, or be he dead",
                "I'll grind his bones to make my bread",
            ]
        }

        expected_task_seq = [
            "task1",
            "task2",
            "task4",
            "task6",
            "task8",
            "task3",
            "task5",
            "task7",
            "task9",
            "task10",
        ]

        mock_results = [
            None,  # task1
            "Fee fi",  # task2
            "I smell the blood of an English man",  # task4
            None,  # task6
            None,  # task8
            "fo fum",  # task3
            None,  # task5
            "Be alive, or be he dead",  # task7
            None,  # task9
            None,  # task10
        ]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name, expected_task_seq, mock_results=mock_results, expected_output=expected_output
        )

    def test_join_task_transition_on_completed(self):
        wf_name = "join-on-complete"

        self.assert_spec_inspection(wf_name)

        expected_task_seq = ["task1", "task2", "task4", "task3", "task5", "task6", "task7"]

        # The tasks before the join, task3 and task5, both succeeded.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

        # First tasks before the join, task3, failed.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

        # Second tasks before the join, task5, failed.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task3
            statuses.FAILED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

        # Both tasks before the join, task3 and task5, failed.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task3
            statuses.FAILED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

    def test_join_task_transition_on_failed(self):
        wf_name = "join-on-fail"

        self.assert_spec_inspection(wf_name)

        expected_task_seq = ["task1", "task2", "task4", "task3", "task5"]

        # The tasks before the join, task3 and task5, both succeeded.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

        # First task before the join, task3, failed.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task3
            statuses.SUCCEEDED,  # task5
        ]

        conductor = self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.FAILED,
        )

        expected_errors = [
            {
                "message": "Execution failed. See result for details.",
                "type": "error",
                "task_id": "task3",
            },
            {
                "message": (
                    'UnreachableJoinError: The join task|route "task6|0" '
                    "is partially satisfied but unreachable."
                ),
                "type": "error",
                "route": 0,
                "task_id": "task6",
            },
        ]

        self.assertListEqual(conductor.errors, expected_errors)

        # Second task before the join, task5, failed.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task3
            statuses.FAILED,  # task5
        ]

        conductor = self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_workflow_status=statuses.FAILED,
        )

        expected_errors = [
            {
                "message": "Execution failed. See result for details.",
                "type": "error",
                "task_id": "task5",
            },
            {
                "message": (
                    'UnreachableJoinError: The join task|route "task6|0" '
                    "is partially satisfied but unreachable."
                ),
                "type": "error",
                "route": 0,
                "task_id": "task6",
            },
        ]

        self.assertListEqual(conductor.errors, expected_errors)

        # The tasks before the join, task3 and task5, both failed.
        expected_task_seq = ["task1", "task2", "task4", "task3", "task5", "task6", "task7"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task3
            statuses.FAILED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        conductor = self.assert_conducting_sequences(
            wf_name, expected_task_seq, mock_statuses=mock_statuses
        )

    def test_join_all_complex(self):
        # In this workflow, both tasks task3 and task5 join all to task6. There
        # are multiple task transition with different criteria (succeeded,
        # failed, and completed) defined at each task. All the task transition
        # specify task6 as the next task. Previously, join all requires all
        # inbound task transition to pass criteria. So for cases where tasks
        # defined multiple task transition with different criteria to the same
        # join task, the join all will never be satisfied. This is now changed
        # such that as long as there is at least one inbound task transition
        # from the same task that meets criteria, then the join condition from
        # that inbound task is considered satisfied.
        wf_name = "join-all-complex"

        self.assert_spec_inspection(wf_name)

        expected_task_seq = ["task1", "task2", "task4", "task3", "task5", "task6", "task7"]

        # Both branch 1 and 2 succeeded.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

        # Both branch 1 and 2 failed and triggering a different
        # set of task transition but will still join on task6.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task3
            statuses.FAILED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

        # One of the branches failed and the other succeeded. Each
        # of the task will trigger different criteria in the task
        # transition but will still join on task6.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

        # One of the branches succeeded and the other failed. Each
        # of the task will trigger different criteria in the task
        # transition but will still join on task6.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task3
            statuses.FAILED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

        # Both branch 1 and 2 failed but trigger the set of task
        # transition that will pass on the failed criteria.
        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task3
            statuses.FAILED,  # task5
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

    def test_join_count_complex(self):
        # In this workflow, the task task3, task5, and task7 join on task8.
        # There are multiple task transition with different criteria (succeeded,
        # failed, and completed) defined at each task. All the task transition
        # specify task8 as the next task. Previously, join count tallies any
        # inbound task transition to pass criteria. This is now changed such
        # that the count is on the inbound task and not on task transition.
        wf_name = "join-count-complex"

        self.assert_spec_inspection(wf_name)

        # All branches succeeded.
        expected_task_seq = [
            "task1",
            "task2",
            "task4",
            "task6",
            "task3",
            "task5",
            "task7",
            "task8",
            "task9",
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.SUCCEEDED,  # task7
            statuses.SUCCEEDED,  # task8
            statuses.SUCCEEDED,  # task9
        ]

        self.assert_conducting_sequences(wf_name, expected_task_seq, mock_statuses=mock_statuses)

        # One of three branches failed. Branch 3 failed.
        expected_task_seq = [
            "task1",
            "task2",
            "task4",
            "task6",
            ("noop", 1),
            "task3",
            "task5",
            "task8",
            "task9",
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task5
            statuses.SUCCEEDED,  # task8
            statuses.SUCCEEDED,  # task9
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task6__t0"]],
        )

        # One of three branches failed. Branch 2 failed.
        expected_task_seq = [
            "task1",
            "task2",
            "task4",
            ("noop", 1),
            "task6",
            "task3",
            "task7",
            "task8",
            "task9",
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.FAILED,  # task4
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task7
            statuses.SUCCEEDED,  # task8
            statuses.SUCCEEDED,  # task9
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task4__t0"]],
        )

        # One of three branches failed. Branch 1 failed.
        expected_task_seq = [
            "task1",
            "task2",
            ("noop", 1),
            "task4",
            "task6",
            "task5",
            "task7",
            "task8",
            "task9",
        ]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.FAILED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task3
            statuses.SUCCEEDED,  # task7
            statuses.SUCCEEDED,  # task8
            statuses.SUCCEEDED,  # task9
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task2__t0"]],
        )

        # Two of three branches failed. Branch 2 and 3 failed.
        expected_task_seq = ["task1", "task2", "task4", ("noop", 1), "task6", ("noop", 2), "task3"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.SUCCEEDED,  # task2
            statuses.FAILED,  # task4
            statuses.FAILED,  # task6
            statuses.SUCCEEDED,  # task3
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task4__t0"], ["task6__t0"]],
            expected_workflow_status=statuses.FAILED,
        )

        # Two of three branches failed. Branch 1 and 2 failed.
        expected_task_seq = ["task1", "task2", ("noop", 1), "task4", ("noop", 2), "task6", "task7"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.FAILED,  # task2
            statuses.FAILED,  # task4
            statuses.SUCCEEDED,  # task6
            statuses.SUCCEEDED,  # task7
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task2__t0"], ["task4__t0"]],
            expected_workflow_status=statuses.FAILED,
        )

        # Two of three branches failed. Branch 1 and 3 failed.
        expected_task_seq = ["task1", "task2", ("noop", 1), "task4", "task6", ("noop", 2), "task5"]

        mock_statuses = [
            statuses.SUCCEEDED,  # task1
            statuses.FAILED,  # task2
            statuses.SUCCEEDED,  # task4
            statuses.FAILED,  # task6
            statuses.SUCCEEDED,  # task5
        ]

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            mock_statuses=mock_statuses,
            expected_routes=[[], ["task2__t0"], ["task6__t0"]],
            expected_workflow_status=statuses.FAILED,
        )



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import yaml

from orquesta.tests.unit.specs import base as test_specs


class SpecTest(unittest.TestCase):
    def setUp(self):
        super(SpecTest, self).setUp()
        self.maxDiff = None

    def test_spec_init_arg_bad_yaml(self):
        self.assertRaises(ValueError, test_specs.MockSpec, "foobar")

    def test_spec_init_yaml(self):
        spec = """
        name: mock
        version: '1.0'
        description: This is a mock spec.
        inputs:
            - x
            - y: polo
        vars:
            var1: foobar
            var2: <% ctx().x %>
            var3: <% ctx().y %>
        attr1: foobar
        attr2:
            macro: polo
        attr3:
            - <% ctx().var1 %>
        attr5:
            attr1:
                attr1: <% ctx().var2 %> <% ctx().var3 %>
        """

        spec_obj = test_specs.MockSpec(spec)
        spec_dict = yaml.safe_load(spec)

        self.assertDictEqual(spec_obj.spec, spec_dict)
        self.assertEqual(spec_obj.name, spec_dict["name"])
        self.assertEqual(spec_obj.version, spec_dict["version"])
        self.assertEqual(spec_obj.description, spec_dict["description"])

        self.assertEqual(spec_obj.attr1, spec_dict["attr1"])
        self.assertDictEqual(spec_obj.attr2, spec_dict["attr2"])
        self.assertListEqual(spec_obj.attr3, spec_dict["attr3"])

        self.assertIsInstance(spec_obj.attr5, test_specs.MockJointSpec)
        self.assertIsInstance(spec_obj.attr5.attr1, test_specs.MockLeafSpec)

        self.assertEqual(spec_obj.attr5.attr1.attr1, spec_dict["attr5"]["attr1"]["attr1"])

        self.assertRaises(AttributeError, getattr, spec_obj, "attr9")

    def test_spec_valid_yaml(self):
        spec = """
        name: mock
        version: '1.0'
        description: This is a mock spec.
        inputs:
            - x
            - y: polo
        vars:
            var1: foobar
            var2: <% ctx().x %>
            var3: <% ctx().y %>
        attr1: foobar
        attr2:
            macro: polo
        attr3:
            - <% ctx().var1 %>
        attr5:
            attr1:
                attr1: <% ctx().var2 %> <% ctx().var3 %>
        """

        spec_obj = test_specs.MockSpec(spec)

        self.assertDictEqual(spec_obj.spec, yaml.safe_load(spec))
        self.assertDictEqual(spec_obj.inspect(), {})



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime
import logging
import six
import ujson

from orquesta.utils import date as date_util


LOG = logging.getLogger(__name__)


SERIALIZABLE_TYPES = (
    six.string_types,
    bool,
    datetime.datetime,
    dict,
    float,
    int,
    list,
)


def serialize(obj):
    doc = {}

    for k, v in six.iteritems(obj.__dict__):
        if isinstance(v, SERIALIZABLE_TYPES):
            doc[k] = date_util.format(v) if date_util.valid(v) else v

    return doc


def deserialize(obj_type, data):
    obj = obj_type()

    for k, v in six.iteritems(data):
        if isinstance(v, SERIALIZABLE_TYPES):
            v = date_util.parse(v) if date_util.valid(v) else v
            setattr(obj, k, v)

    return obj


def deepcopy(value):
    # NOTE: ujson round-trip is up to 10 times faster on smaller and larger dicts compared
    # to copy.deepcopy(), but it has some edge cases with non-simple types such as datetimes.
    try:
        value = ujson.loads(ujson.dumps(value))  # pylint: disable=no-member
    except (OverflowError, ValueError, TypeError):
        # NOTE: ujson doesn't support 5 or 6 bytes utf-8 sequences or
        # non-JSON serializable object so we fallback to copy.deepcopy.
        value = copy.deepcopy(value)

    return value



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta import events
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorWithItemsFailureTest(test_base.WorkflowConductorWithItemsTest):
    def test_fail_one_and_only_item(self):
        wf_def = """
        version: 1.0
        vars:
          - xs:
              - fee
        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
        ]

        mock_ac_ex_statuses = [statuses.FAILED]
        expected_task_statuses = [statuses.FAILED]
        expected_workflow_statuses = [statuses.FAILED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging. This is intentional so the with items
        # task can be rerun partially for failed items or items that hasn't been run.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow failed.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

    def test_failed_item_task_dormant_other_incomplete(self):
        wf_def = """
        version: 1.0
        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum
        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED, statuses.FAILED]
        expected_task_statuses = [statuses.RUNNING, statuses.FAILED]
        expected_workflow_statuses = [statuses.RUNNING, statuses.FAILED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging. This is intentional so the with items
        # task can be rerun partially for failed items or items that hasn't been run.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow failed.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

    def test_failed_item_task_dormant_other_failed(self):
        wf_def = """
        version: 1.0
        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum
        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED, statuses.FAILED, statuses.FAILED]
        expected_task_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.FAILED]
        expected_workflow_statuses = [statuses.RUNNING, statuses.RUNNING, statuses.FAILED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging. This is intentional so the with items
        # task can be rerun partially for failed items or items that hasn't been run.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow failed.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

    def test_failed_item_task_active(self):
        wf_def = """
        version: 1.0
        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum
        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [
            statuses.SUCCEEDED,
            statuses.FAILED,
            statuses.SUCCEEDED,
            statuses.SUCCEEDED,
        ]

        expected_task_statuses = [statuses.RUNNING] * 3 + [statuses.FAILED]

        expected_workflow_statuses = [statuses.RUNNING] * 3 + [statuses.FAILED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is not removed from staging. This is intentional so the with items
        # task can be rerun partially for failed items or items that hasn't been run.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow failed.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

    def test_failed_item_task_dormant_with_concurrency(self):
        wf_def = """
        version: 1.0
        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum
        tasks:
          task1:
            with:
              items: <% ctx(xs) %>
              concurrency: 2
            action: core.echo message=<% item() %>
        """

        concurrency = 2

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED, statuses.FAILED]
        expected_task_statuses = [statuses.RUNNING, statuses.FAILED]
        expected_workflow_statuses = [statuses.RUNNING, statuses.FAILED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
            concurrency=concurrency,
        )

        # Assert the task is not removed from staging. This is intentional so the with items
        # task can be rerun partially for failed items or items that hasn't been run.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow failed.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

    def test_failed_item_task_active_with_concurrency(self):
        wf_def = """
        version: 1.0
        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum
        tasks:
          task1:
            with:
              items: <% ctx(xs) %>
              concurrency: 2
            action: core.echo message=<% item() %>
        """

        concurrency = 2

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [
            statuses.SUCCEEDED,
            statuses.FAILED,
            statuses.SUCCEEDED,
            statuses.SUCCEEDED,
        ]

        expected_task_statuses = [statuses.RUNNING] * 3 + [statuses.FAILED]

        expected_workflow_statuses = [statuses.RUNNING] * 3 + [statuses.FAILED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
            concurrency=concurrency,
        )

        # Assert the task is not removed from staging. This is intentional so the with items
        # task can be rerun partially for failed items or items that hasn't been run.
        self.assertIsNotNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow failed.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)

    def test_failed_item_with_task_remediation(self):
        wf_def = """
        version: 1.0
        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum
        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - do: task2
          task2:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [
            statuses.SUCCEEDED,
            statuses.FAILED,
            statuses.SUCCEEDED,
            statuses.SUCCEEDED,
        ]

        expected_task_statuses = [statuses.RUNNING] * 3 + [statuses.FAILED]

        expected_workflow_statuses = [statuses.RUNNING] * 4

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        current_task = conductor.get_next_tasks()[0]
        current_task_id = current_task["id"]
        current_task_route = current_task["route"]
        ac_ex_event = events.ActionExecutionEvent(statuses.RUNNING)
        conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)
        ac_ex_event = events.ActionExecutionEvent(statuses.SUCCEEDED)
        conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)

        # Assert there are only two tasks executed.
        self.assertEqual(len(conductor.workflow_state.sequence), 2)

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_failed_all_items_with_task_remediation(self):
        wf_def = """
        version: 1.0
        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum
        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - do: task2
          task2:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [
            statuses.FAILED,
            statuses.FAILED,
            statuses.FAILED,
            statuses.FAILED,
        ]

        expected_task_statuses = [statuses.RUNNING] * 3 + [statuses.FAILED]

        expected_workflow_statuses = [statuses.RUNNING] * 4

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        current_task = conductor.get_next_tasks()[0]
        current_task_id = current_task["id"]
        current_task_route = current_task["route"]
        ac_ex_event = events.ActionExecutionEvent(statuses.RUNNING)
        conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)
        ac_ex_event = events.ActionExecutionEvent(statuses.SUCCEEDED)
        conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)

        # Assert there are only two tasks executed.
        self.assertEqual(len(conductor.workflow_state.sequence), 2)

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_failed_item_task_dormant_with_concurrency_and_task_remediation(self):
        wf_def = """
        version: 1.0
        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum
        tasks:
          task1:
            with:
              items: <% ctx(xs) %>
              concurrency: 2
            action: core.echo message=<% item() %>
            next:
              - do: task2
          task2:
            action: core.noop
        """

        concurrency = 2

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED, statuses.FAILED]
        expected_task_statuses = [statuses.RUNNING, statuses.FAILED]
        expected_workflow_statuses = [statuses.RUNNING] * 4

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
            concurrency=concurrency,
        )

        current_task = conductor.get_next_tasks()[0]
        current_task_id = current_task["id"]
        current_task_route = current_task["route"]
        ac_ex_event = events.ActionExecutionEvent(statuses.RUNNING)
        conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)
        ac_ex_event = events.ActionExecutionEvent(statuses.SUCCEEDED)
        conductor.update_task_state(current_task_id, current_task_route, ac_ex_event)

        # Assert there are only two tasks executed.
        self.assertEqual(len(conductor.workflow_state.sequence), 2)

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_task_cycle(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          init:
            next:
              - do: task1
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - when: <% failed() %>
                do: task1
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Get pass the init task, required for bootstrapping self looping task..
        self.forward_task_statuses(conductor, "init", [statuses.RUNNING, statuses.SUCCEEDED])

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED, statuses.FAILED]
        expected_task_statuses = [statuses.RUNNING, statuses.FAILED]
        expected_workflow_statuses = [statuses.RUNNING] * 2

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is reset in staging.
        staged_task = conductor.workflow_state.get_staged_task(task_name, task_route)
        self.assertIsNotNone(staged_task)
        self.assertNotIn("items", staged_task)

        # Assert the workflow is still running.
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        # Mock the second task execution.
        mock_ac_ex_statuses = [statuses.SUCCEEDED] * 4
        expected_task_statuses = [statuses.RUNNING] * 3 + [statuses.SUCCEEDED]
        expected_workflow_statuses = [statuses.RUNNING] * 3 + [statuses.SUCCEEDED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
